// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;                                               // for Math
using System.Collections.Generic;                           // for List(2), Dictionary(1)
using System.Linq;
using System.Text;                                          // for StringBuilder(1)
using System.Text.RegularExpressions;                       // for Regex

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- Endeme -->
    /// <summary>
    ///      An endeme is a data structure for storing information, it is an instance of a
    ///      permutable combinable list of characteristics and works a bit like an ordered bitwise
    ///      enum, The first letter is generally considered position 0
    /// </summary>
    /// <remarks>
    ///      Data vs Information vs Talent
    /// 
    ///      The reason it is so hard to define the distinction between data and information is
    ///      that the distinction is partially based on the talent of the reader. The general (very loose)
    ///      definition of information is that 'information is data that has meaning to people'.
    ///      Different people have different talent for interpreting data and are therefore able to process data different
    ///      ways. So different people are able to understand different data sets. To them it is
    ///      information, to others it has no meaning and therefore it is only data.
    /// 
    ///      For example someone with high levels of inductive reasoning aptitude will be able to
    ///      process larger amounts of more diffuse data
    ///      into meaning that will a person with low inductive reasoning. Someone who is very
    ///      good at deduction will be able to take data in the form of logic and basic principles
    ///      and apply them to an understanding rooted in concrete systems where another person would
    ///      not find meaning in them. Someone good at recognizing particular pattern commonalities
    ///      might find that certain data show an obvious match whereas other people would not see
    ///      the match. In each case a particular data set would be information for some people and
    ///      data for others.
    /// 
    ///      One could define information as data that is understandable to most people, or as data
    ///      understandable to a few people, but then where do you draw the lines?
    /// 
    ///      I have developed a data structure which I call an endeme, which I believe stores
    ///      information rather than data. Essentially, it provides people a list of 'factors'
    ///      presented from strongest to weakest in order, to describe something. Clearly,
    ///      this might be more understandable to some people than others.
    ///      In a more basic sense it is information rather than data because the
    ///      strengths of the factors are defined relative to each other rather than in hard numbers.
    /// 
    ///      So I have snuck in here a second definition of information vs data. Information is fuzzy
    ///      and data is precise. In an array of data, you might have a list of 22 people with
    ///      precise ages identified for each. In an endeme, you would have the same set of people
    ///      ordered by age without giving specific ages, but only given perhaps an average age for 
    ///      the group. Whe it comes to ages, basic data is probably more useful. When it comes to
    ///      combinations of concepts, endemes would be more useful because concepts tend to be
    ///      fuzzy anyway.
    ///      
    ///      Another way of defining information is something that has meaning. Endemes store information
    ///      because they store the relative importance of things with each other, and importance is a form
    ///      of meaning, perhaps the most important form of meaning. In other words what is most important
    ///      to someone is what has the most meanin to them.
    /// </remarks>
    public class Endeme
    {
        // ----------------------------------------------------------------------------------------
        //  Members
        //    Because there is derived information, the authoritative, source and the derived source
        //    both have to remain private, because you have to change both at the same time under control
        // ----------------------------------------------------------------------------------------
        private  List<char>            _string;                                        // _string is the authoritative information source
        private  Dictionary<char, int> _index;                                         // _index is the derived information source
        internal int                   Size    { get { return _size; } }  private int        _size = 22; // 17 is a 'magic number', it produces no negtive raw matches for a 22 char endeme set - I should probably switch to 26
        public   EndemeSet             Set     { get { return _set;  } }  private EndemeSet  _set;
        public   EndemeQuantification  Quant   { get; set; }


        // ----------------------------------------------------------------------------------------
        //  Constructors
        // ----------------------------------------------------------------------------------------
        public Endeme(                                          ) { Init(null, null                ); } //     _isEmpty = true);  }
        public Endeme(EndemeSet set                             ) { Init(set , null                ); } //     _isEmpty = true);  }
        public Endeme(EndemeSet set, Dictionary<char,int> index ) { Init(set , IndexToString(index)); } //     _isEmpty = false); }
        public Endeme(EndemeSet set,               string endeme) { Init(set , endeme              ); } //     _isEmpty = false); }
        public Endeme(                             string endeme) { Init(null, endeme              ); } //     _isEmpty = false); }
        public Endeme(char[] arr, EndemeOrderingCreationType eot) { _set = null; BuildString(arr, eot); BuildIndex(); Quant = new EndemeQuantification(); } //_isEmpty = false; }


        // ----------------------------------------------------------------------------------------
        //  Short methods, properties, accessors, operators and casting (implicit operators)
        // ----------------------------------------------------------------------------------------
        public char   this[int  i]    { get { if (_string.Count > i)     return _string[i];    return '\0';   } }
        public int    this[char c]    { get { if (_index.ContainsKey(c)) return _index[c];     return -1;     } }
        public int    Count           { get { if (_string != null)       return _string.Count; return -1;     } } /// <summary>Returns the number of actual letters in the endeme (may be less than the number in the endeme set)</summary>
        public string FirstTwo        { get {                            return First(2);                     } } /// <summary>Returns the set name and theendeme string delimited by a colon</summary>
        public string FullDesignation { get {                            return (SetName + ":" + ToString()); } }
        public bool   IsEmpty         { get {                            return (_string == null || _string.Count == 0); } }
        public string SetName         { get { if (_set    != null)       return _set.Name;     return "";     } } /// <summary>Returns the array of letters</summary>
        public char[] ToCharArray()         {                            return _string.ToArray();            }   /// <summary>Adds a value to the quantification of an endeme characteristic</summary>
        public void   AddToLetter(char c, int num) { if ('A' <= c && c <= 'Z') Quant.Raw[c] += num;           }  
        // ----------------------------------------------------------------------------------------
        public static implicit operator   Endeme     (char   cha) { return new Endeme(cha.ToString());        }  
        public static implicit operator   Endeme     (string str) { return new Endeme(str);                   }   ///<summary>WARNING - multiplication is not associative, use parentheses when you have more than 2 endemes!</summary>
        public static Endeme   operator *(Endeme lhs, Endeme rhs) { return MultiplyTogether(lhs, rhs);        }   ///<summary>WARNING - addition is not associative, use parentheses when you have more than 2 endemes!</summary>
        public static Endeme   operator +(Endeme lhs, Endeme rhs) { return AddTogether(lhs, rhs);             }  
        public static                     Endeme Empty      { get { return new Endeme(EndemeSet.Empty);       } } /// <summary>Used for Equals() and GetHashCode()</summary>
        public        override            int    GetHashCode()    { return FullDesignation.GetHashCode();     }


        // --------------------------------------------------------------------------------------
        /// <!-- Add -->
        /// <summary>
        ///      Adds a character to the index and the authoritative source _string
        /// </summary>
        /// <param name="c"></param>
        /// <param name="i"></param>
        private void Add(char c, int i)
        {
            if (_index == null)
                _index = new Dictionary<char,int>(23);
            _string.Add(c);
            _index.Add(c, _size - i);
        }


        // --------------------------------------------------------------------------------------
        /// <!-- AddTogether -->
        /// <summary>
        ///      Overloads the + operator by merging the endemes using addition and linear series
        /// </summary>
        /// <param name="lhs"></param>
        /// <param name="rhs"></param>
        /// <returns></returns>
        private static Endeme AddTogether(Endeme lhs, Endeme rhs)
        {
            if (lhs == null && rhs == null) return new Endeme();
            if (lhs == null) return rhs.Copy();
            if (rhs == null) return lhs.Copy();


            Endeme endeme = new Endeme(lhs.Set);


            // --------------------------------------------------------------------------
            //  If they are the same, just return a copy
            // --------------------------------------------------------------------------
            if (lhs.Equals(rhs))
                endeme = lhs.Copy();
            else
            {
                // ----------------------------------------------------------------------
                //  Prepare the objects
                // ----------------------------------------------------------------------
                Endeme left = lhs.Copy(); left.GenerateCalc_linear();
                Endeme right = rhs.Copy(); right.GenerateCalc_linear();
                char[] chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ".ToCharArray();


                // ----------------------------------------------------------------------
                //  Do the calculation
                // ----------------------------------------------------------------------
                for (int i = 0; i < lhs.Size; ++i)
                {
                    char c = chars[i];
                    endeme.Quant.Raw[c] = left.Quant.Calc[c] + right.Quant.Calc[c];
                }
                endeme.CookEndeme();
            }


            return endeme;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ArrayByFrequency -->
        /// <summary>
        ///      Orders a character array by frequency from most to least frequent
        /// </summary>
        /// <param name="array"></param>
        /// <returns></returns>
        private static char[] ArrayByFrequency(char[] array)
        {
            // --------------------------------------------------------------------------
            //  Get the counts for each character into a dictionary
            // --------------------------------------------------------------------------
            Dictionary<char, int> counter = new Dictionary<char, int>();
            for (int i = 0; i < array.Length; ++i)
            {
                char c = array[i];
                if (!counter.ContainsKey(c)) counter.Add(c, 1);
                else counter[c]++;
            }


            // --------------------------------------------------------------------------
            //  Create the new array by getting the number of characters
            // --------------------------------------------------------------------------
            var sortedList = (from d in counter orderby d.Value descending select d);
            int count = 0;
            foreach (var item in sortedList)
                count++;
            char[] array2 = new char[count];


            // --------------------------------------------------------------------------
            //  Fill the new array
            // --------------------------------------------------------------------------
            int j = 0;
            foreach (var item in sortedList)
            {
                array2[j] = item.Key;
                j++;
            }


            return array2;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- BuildIndex -->
        /// <summary>
        ///      Refreshes the endeme index
        /// </summary>
        /// <remarks>
        ///      Treat as private, internal only for testing purposes
        /// </remarks>
        internal void BuildIndex()
        {
            int count = _string.Count;
            _index = new Dictionary<char,int>(count);
            for (int i = 0; i < count; ++i)
                _index.Add(_string[i], _size - i);
        }


        // --------------------------------------------------------------------------------------
        /// <!-- BuildString -->
        /// <summary>
        ///      Adds each unique letter to the endeme
        /// </summary>
        /// <remarks>
        ///      Treat as private, internal only for testing purposes
        /// </remarks>
        /// <param name="endeme"></param>
        internal void BuildString(string endeme)
        {
            // --------------------------------------------------------------------------
            //  Build an endeme string out of chars A-Z only:
            // --------------------------------------------------------------------------
            if (string.IsNullOrEmpty(endeme))
                endeme = "";
            char[] cha = Regex.Replace(endeme.ToUpper(), "[^A-Z]", "").ToCharArray();
            if (_set == null || _set.Count == 0)
                BuildString(cha);
            else
                BuildString(cha, _set);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- BuildString -->
        /// <summary>
        ///      Without a set, add all the letters (once only for each letter)
        /// </summary>
        /// <param name="cha"></param>
        private void BuildString(char[] cha)
        {
            _string = new List<char>();
            _index = new Dictionary<char, int>();
            int count = cha.Length;
            for (int i = 0; i < count; ++i)
                if (!Contains(cha[i]))
                    Add(cha[i], i);
        }
        private void BuildString(char[] arr, EndemeOrderingCreationType type)
        {
            if (type == EndemeOrderingCreationType.ByCharacterFrequency) 
                arr = ArrayByFrequency(arr);
            BuildString(arr);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- BuildString -->
        /// <summary>
        ///      With a set, add only the letters in the set (once each)
        /// </summary>
        /// <remarks>
        ///      Treat as private, internal only for testing purposes
        /// </remarks>
        /// <param name="endeme"></param>
        /// <param name="set"></param>
        internal void BuildString(char[] cha, EndemeSet set)
        {
            _string = new List<char>();
            _index = new Dictionary<char, int>();
            int count = cha.Length;
            for (int i = 0; i < count; ++i)
                if (!Contains(cha[i]) && set.Contains(cha[i]))
                    Add(cha[i], i);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- CollateIndex -->
        /// <summary>
        ///      Returns a string containing the numbers in the index offset by a particular value
        /// </summary>
        /// <param name="join"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public string CollateIndex(string join, int offset)
        {
            offset -= _size;
            StringBuilder str = new StringBuilder();
            List<char> keys = KeysSorted;
            string delim = "";
            foreach (char key in keys)
            {
                str.Append(delim + (_index[key] + offset));
                delim = join;
            }
            return str.ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- CookEndeme -->
        /// <summary>
        ///      Takes the raw assembled data and cooks it into the endeme
        /// </summary>
        public void CookEndeme()
        {
            var sortedList = (from d in Quant.Raw where d.Value != 0 orderby d.Value descending select d);
            string endeme = "";
            foreach (var item in sortedList) endeme += item.Key;


            _index = new Dictionary<char, int>(endeme.Length);
            BuildString(endeme);
            BuildIndex();
            //_isEmpty = false;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Contains -->
        /// <summary>
        /// 
        /// </summary>
        /// <summary>
        ///      This is killing our performance (49.264%)
        ///      I must use a hash to fix this
        /// </summary>
        /// <param name="cha"></param>
        /// <returns></returns>
        public bool Contains(char cha)
        {
            if (_index == null)
                return _string.Contains(cha);   // this is a sequential search and therefore very slow
            else
                return _index.ContainsKey(cha); // you want to use the index if it exists
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Copy -->
        /// <summary>
        ///      Returns a copy of itself
        /// </summary>
        /// <returns></returns>
        public Endeme Copy()
        {
            Endeme e = new Endeme();
            e._string  = this.ToList();  // this is a copy
            //e._isEmpty = _isEmpty;       // this is a copy
            e._set     = Set;            // do not copy the set, just include the reference
            e.Quant    = Quant;          // do not copy, just include the reference for now
            e.BuildIndex();              // this is a copy
            return e;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Equals -->
        /// <summary>
        ///      Checks to see if the core parts (string and set) of the endeme are identical
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public override bool Equals(object obj)
        {
            if (obj == null)  return false;
            if (obj == this)  return true;
            if (! (obj is Endeme))  return false;
            return (this.FullDesignation == ((Endeme)obj).FullDesignation);
        }
        // ----------------------------------------------------------------------------------------
        //  This stuff does not work because it causes infinite loops:
        // ----------------------------------------------------------------------------------------
        //public static bool operator ==(Endeme lhs, Endeme rhs)
        //{
        //    if (rhs == null)
        //        return false;
        //    return (lhs.FullDesignation == rhs.FullDesignation);
        //}
        //public static bool operator !=(Endeme lhs, Endeme rhs)
        //{
        //    if (rhs == null)
        //        return true;
        //    return (lhs.FullDesignation != rhs.FullDesignation);
        //}


        // ----------------------------------------------------------------------------------------
        /// <!-- FormatFourEndemes -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="e1"></param>
        /// <param name="s1"></param>
        /// <param name="e2"></param>
        /// <param name="e3"></param>
        /// <param name="e4"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        public static string FormatFourEndemes(Endeme e1, string s1, Endeme e2, Endeme e3, Endeme e4, int width)
        {
            StringBuilder str = new StringBuilder();
            str.Append("|" + SetLength(width, " " + e1.SetName + ": " + s1)
                + "| " + e2.SetName.PadRight(width-1, ' ')
                + "| " + e3.SetName.PadRight(width-1, ' ')
                + "| " + e4.SetName.PadRight(width-1, ' ')
                + "|");
            str.Append("\r\n");
            str.Append("| ".PadRight(width, '-') + " "
                + "| ".PadRight(width, '-') + " "
                + "| ".PadRight(width, '-') + " "
                + "| ".PadRight(width, '-') + " "
                + "|");
            string delim = "\r\n";
            for (int i = 0; i < 22; ++i)
            {
                str.Append(delim);
                str.Append("|" + FormatEndemeLeaf(e1, i, width));
                str.Append("|" + FormatEndemeLeaf(e2, i, width));
                str.Append("|" + FormatEndemeLeaf(e3, i, width));
                str.Append("|" + FormatEndemeLeaf(e4, i, width));
                str.Append("|");
                delim = "\r\n";
            }
            return str.ToString();
        }
        private static string FormatEndemeLeaf(Endeme e, int num, int width)
        {
            char[] alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ".ToCharArray();
            char[] c = e.ToCharArray();
            string str = "";
            if (num < c.Length)
                str = " " + c[num] + " |"
                    + " " + e.Index(alphabet[num]).ToString().PadLeft(2,' ')
                    + " " + e.Set.Characteristic[alphabet[num]].Display;
            return SetLength(width, str);
        }


        // --------------------------------------------------------------------------------------
        /// <!-- First -->
        /// <summary>
        ///      Returns the first n items as a string of letters
        /// </summary>
        /// <param name="num"></param>
        /// <returns></returns>
        public string First(int num)
        {
            char[] arr = _string.ToArray();
            string str = "";
            for (int i = 0; i < arr.Length && i < num; ++i)
                str += arr[i];
            return str;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- GenerateCalc_geometric -->
        /// <summary>
        ///      Generates the Calc values from the endeme string using a geometric slope
        /// </summary>
        /// <param name="geometric">usually between 0.0 and 1.0, 0.93 is good</param>
        public void GenerateCalc_geometric(double geometric)
        {
            char[] letters = _string.ToArray();
            double num = 100.0;
            foreach (char c in letters)
            {
                Quant.Calc[c] = num;
                num = num * geometric;
            }
        }


        // --------------------------------------------------------------------------------------
        /// <!-- GenerateCalc_linear -->
        /// <summary>
        ///      Generates the Calc values from the endeme string using a linear slope
        /// </summary>
        /// <param name="slope">usually negative, -4.55 is a good number</param>
        public void GenerateCalc_linear(double slope)
        {
            char[] letters = ToString().ToCharArray();
            double num = 100.0;
            foreach (char c in letters)
            {
                Quant.Calc[c] = num;
                num = num + slope;
            }
        }
        public void GenerateCalc_linear() { GenerateCalc_linear(-100.0 / _size); }


        // --------------------------------------------------------------------------------------
        /// <!-- GenerateRaw_geometric -->
        /// <summary>
        ///      Generates the raw values from the endeme string using a geometric slope
        /// </summary>
        /// <param name="geometric">usually between 0.0 and 1.0, 0.93 is good</param>
        public void GenerateRaw_geometric(double geometric)
        {
            char[] letters = _string.ToArray();
            double num = 100.0;
            foreach (char c in letters)
            {
                Quant.Raw[c] = num;
                num = num * geometric;
            }
        }


        // --------------------------------------------------------------------------------------
        /// <!-- GenerateRaw_linear -->
        /// <summary>
        ///      Generates the raw values from the endeme string using a linear slope
        /// </summary>
        /// <param name="slope">usually negative, -4.55 is a good number</param>
        public void GenerateRaw_linear(double slope)
        {
            char[] letters = _string.ToArray();
            double num = 100.0;
            foreach (char c in letters)
            {
                Quant.Raw[c] = num;
                num = num + slope;
            }
        }
        public void GenerateRaw_linear() { GenerateRaw_linear(-100.0 / _size); }


        // --------------------------------------------------------------------------------------
        /// <!-- Index -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="cha"></param>
        /// <returns></returns>
        public int Index(char cha)
        {
            if      (IsEmpty)                 return -99;
            else if (_index.ContainsKey(cha)) return _index[cha];
            else                              return -99;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- IndexToString -->
        /// <summary>
        ///      Converts a dictionary of character positions into an Endeme
        /// </summary>
        /// <param name="positions"></param>
        /// <returns></returns>
        private static string IndexToString(Dictionary<char,int> positions)
        {
            List<char> e = new List<char>(positions.Count);
            for (int i = 0; i < 32 + positions.Count; ++i) e.Add(' ');
            foreach (char cha in positions.Keys)
            {
                int pos = 32 - positions[cha];
                e[pos] = cha;
            }
            string endeme = new string(e.ToArray());
            endeme = Regex.Replace(endeme, " ", "");
            return endeme;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- Init -->
        /// <summary>
        ///      Initializes an Endeme object
        /// </summary>
        /// <param name="set"></param>
        /// <param name="endeme"></param>
        /// <param name="isEmpty"></param>
        private void Init(EndemeSet set, string endeme)
        {
            // --------------------------------------------------------------------------
            //  Put the string together
            // --------------------------------------------------------------------------
            _set = set;
            if (endeme == null && set != null)
                BuildString(set.Letters);
            else
                BuildString(endeme);


            // --------------------------------------------------------------------------
            //  Calculate indexing and such
            // --------------------------------------------------------------------------
            BuildIndex();
            Quant = new EndemeQuantification();
            if (_set != null && _set.Count > 0)
                _size = _set.Count;
            else _size = 22;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- KeysSorted -->
        /// <summary>
        ///      Returns an alphabetically sorted list of the letters in the endeme
        /// </summary>
        /// <returns></returns>
        private List<char> KeysSorted { get
        {
            List<char> keys = new List<char>(_index.Keys.Count);
            foreach (char key in _index.Keys)
                keys.Add(key);
            keys.Sort();
            return keys;
        } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Labels -->
        /// <summary>
        ///      Returns a string of the first n characteristics comma delimited
        /// </summary>
        /// <param name="num">number of characteristics labels to show</param>
        /// <returns></returns>
        public string Labels(int num)
        {
            return _set.Labels(First(num));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Move -->
        /// <summary>
        ///      Moves characters to the beginning and/or ending of the endeme
        /// </summary>
        /// <param name="toFront">the characters to move to the front</param>
        /// <param name="toBack">the characters to move to the back</param>
        /// <returns></returns>
        public Endeme Move(string toFront, string toBack)
        {
            Endeme temp = this;
            if (!string.IsNullOrEmpty(toFront))
                temp = new Endeme
                    ( Set, toFront + Regex.Replace(temp.ToString(), "[" + toFront + "]", ""));
            if (!string.IsNullOrEmpty(toBack))
                temp = new Endeme
                    ( Set, Regex.Replace(temp.ToString(), "[" + toBack  + "]", "") + toBack);
            this._string = temp._string;
            this._index  = temp._index;
            return this;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- MultiplyTogether -->
        /// <summary>
        ///      Overloads the * operator by merging the endemes using multiplication and geometric series
        /// </summary>
        /// <param name="lhs"></param>
        /// <param name="rhs"></param>
        /// <returns></returns>
        private static Endeme MultiplyTogether(Endeme lhs, Endeme rhs)
        {
            Endeme endeme = new Endeme(lhs.Set);


            // --------------------------------------------------------------------------
            //  If they are the same, just return a copy
            // --------------------------------------------------------------------------
            if (lhs.Equals(rhs))
                endeme = lhs.Copy();
            else
            {
                // ----------------------------------------------------------------------
                //  Prepare the objects
                // ----------------------------------------------------------------------
                Endeme left  = lhs.Copy(); left.GenerateCalc_geometric(0.93);
                Endeme right = rhs.Copy(); right.GenerateCalc_geometric(0.93);
                char[] chars = "ABCDEFGHIJKLMNOPQRSTUV".ToCharArray();


                // ----------------------------------------------------------------------
                //  Do the calculation
                // ----------------------------------------------------------------------
                for (int i = 0; i < lhs.Size; ++i)
                {
                    char c = chars[i];
                    endeme.Quant.Raw[c] = left.Quant.Calc[c] * right.Quant.Calc[c];
                }
                endeme.CookEndeme();
            }


            return endeme;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Reverse -->
        /// <summary>
        ///      Returns the reverse of the endeme
        /// </summary>
        public Endeme Reverse { get
        {
            Endeme e = new Endeme();
            e._string = this.ToList();
            e._string.Reverse();
            e.BuildIndex();
            //e._isEmpty = _isEmpty;
            return e;
        } }


        // ----------------------------------------------------------------------------------------
        /// <!-- SetLength -->
        /// <summary>
        ///      Returns a string exactly as long as specified
        /// </summary>
        /// <remarks>
        ///      This is copied from Core to keep this namespace atomic
        /// </remarks>
        /// <param name="str"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        public static string SetLength(int width, string str)
        {
            if (str == null) str = "";
            str = str.PadRight(width+1, ' ').Substring(0,width);
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Show -->
        /// <summary>
        ///      Returns a string displaying the endeme in the specified width
        /// </summary>
        /// <remarks>
        ///      112-2334-4556-6778-7654-321
        ///         1    3    5    4    2
        /// </remarks>
        /// <param name="width"></param>
        /// <param name="divs"></param>
        /// <param name="divider"></param>
        /// <returns></returns>
        public string Show_complex(int width, int divs, string divider)
        {
            // --------------------------------------------------------------------------
            //  Handle some error conditions
            // --------------------------------------------------------------------------
            if (width <= 0) return "";
            if (IsEmpty || _string == null)
                return divider.PadRight(divs,divider.ToCharArray()[0]).PadRight(width);
            if (divs >= width)
                return "".PadRight(width, divider.ToCharArray()[0]);


            // --------------------------------------------------------------------------
            //  Calculate how many characteristics to show
            // --------------------------------------------------------------------------
            divs = Math.Min(5, divs);
            int num = width - divs * divider.Length;
            if (Set == null)  num = Math.Min(num, _string.Count);
            else              num = Math.Min(num, Set.Count);


            // --------------------------------------------------------------------------
            //  Remove characters from the endeme string as necessary to format
            // --------------------------------------------------------------------------
            if (width <= 8)              return Show_positive(width, divs, num, divider);
            if (num < ToString().Length) return Show_partial(width, divs, num, divider);
            else                         return Show_perfect(width, divs, num, divider);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Show -->
        /// <summary>
        ///      Shows an endeme in groups of three characters
        /// </summary>
        /// <param name="width"></param>
        /// <returns></returns>
        public string Show(int width)
        {
            string endeme = ToString();
            endeme = Regex.Replace(endeme, "^(...)(.)", "$1-$2");
            endeme = Regex.Replace(endeme, "^(...-...)(.)", "$1-$2");
            endeme = Regex.Replace(endeme, "^(...-...-...)(.)", "$1-$2");
            if (endeme.Length > 20)
            {
                endeme = Regex.Replace(endeme, "(.)(...)$", "$1-$2");
                endeme = Regex.Replace(endeme, "(.)(...-...)$", "$1-$2");
                endeme = Regex.Replace(endeme, "(.)(...-...-...)$", "$1-$2");
            }
            else
            {
                endeme = Regex.Replace(endeme, "^(...-...-...-....)(.)", "$1-$2");
                endeme = Regex.Replace(endeme, "^(...-...-...-....-...)(.)", "$1-$2");
                endeme = Regex.Replace(endeme, "^(...-...-...-....-...-...)(.)", "$1-$2");
            }
            while (endeme.Length > width)
            {
                int chop = (int)(0.7 * endeme.Length);
                endeme = endeme.Substring(0, chop-1) + "|" + endeme.Substring(chop + 1);
            }
            return endeme.PadRight(width);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Show_positive -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="width"></param>
        /// <param name="divs"></param>
        /// <param name="num"></param>
        /// <param name="divider"></param>
        /// <returns></returns>
        private string Show_positive(int width, int divs, int num, string divider)
        {
            string endeme = ToString();
            endeme = endeme.Substring(0, width - divs);
            int len = endeme.Length;


            if (divs == 1)
            {
                if      (len >= 8) endeme = Regex.Replace(endeme, "^(.......)", "$1-");
                else if (len >= 3) endeme = Regex.Replace(endeme, "^(...)"    , "$1-");
                else               endeme = endeme + "-";
            }


            if (divs >  1) { if (len >= 10) endeme = Regex.Replace(endeme, "^(.......)"               , "$1-"); else endeme = endeme + "-"; }
            if (divs >= 2) { if (len >=  3) endeme = Regex.Replace(endeme, "^(...)"                   , "$1-"); else endeme = endeme + "-"; }
            if (divs >= 3) { if (len >= 17) endeme = Regex.Replace(endeme, "^(...-....-........)"     , "$1-"); else endeme = endeme + "-"; }
            if (divs >= 4) { if (len >= 20) endeme = Regex.Replace(endeme, "^(...-....-........-....)", "$1-"); else endeme = endeme + "-"; }
            if (divs >= 5) { if (len >= 13) endeme = Regex.Replace(endeme, "^(...-....-....)"         , "$1-"); else endeme = endeme + "-"; }
            return endeme.PadRight(width);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Show_partial -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="width"></param>
        /// <param name="divs"></param>
        /// <param name="divider"></param>
        /// <returns></returns>
        public string Show_partial(int width, int divs, int num, string divider)
        {
            string endeme = ToString();
            int lenA = endeme.Length;


            // --------------------------------------------------------------------------
            //  Construct the initial partial endeme
            // --------------------------------------------------------------------------
            int len1 = (int)(0.5 + 2.0 * num / 3.0);
            int len2 = (int)(0.5 + 1.0 * num / 3.0);
            string part1 = endeme.Substring(0, len1);
            string part2 = endeme.Substring(lenA - len2, len2);
            if (divs > 0) endeme = part1 + "_" + part2;
            else endeme = part1 + part2;


            // --------------------------------------------------------------------------
            //  Insert more dividers
            // --------------------------------------------------------------------------
            int len = endeme.Length;
            if (divs >= 2 && endeme.Length < width)
            {
                if (part1.Length >= 6) endeme = Regex.Replace(endeme, "^(.......)", "$1-");
                else endeme = Regex.Replace(endeme, "_", "_-");
            }
            if (divs >= 3 && endeme.Length < width) endeme = Regex.Replace(endeme, "^(...)", "$1-");
            if (divs >= 4 && endeme.Length < width) endeme = Regex.Replace(endeme, "(...)$", "-$1");
            if (divs >= 5 && endeme.Length < width)
            {
                if (part2.Length >= 7) endeme = Regex.Replace(endeme, "(........)$", "-$1");
                else endeme = Regex.Replace(endeme, "_", "_-");
            }

            endeme = Regex.Replace(endeme, "_", "-");
            endeme = Regex.Replace(endeme, "-", divider);

            return endeme;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Show_perfect -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="width"></param>
        /// <param name="divs"></param>
        /// <param name="divider"></param>
        /// <returns></returns>
        public string Show_perfect(int width, int divs, int num, string divider)
        {
            string endeme = ToString();

            // --------------------------------------------------------------------------
            //  Insert dividers
            // --------------------------------------------------------------------------
            int len = endeme.Length;
            if (divs >= 1 && len >= 10) endeme = Regex.Replace(endeme, "^(.......)", "$1-");
            if (divs >= 2 && len >= 14) endeme = Regex.Replace(endeme, "(.......)$", "-$1");
            if (divs >= 3 && len >=  3) endeme = Regex.Replace(endeme, "^(...)", "$1-");
            if (divs >= 4 && len >=  6) endeme = Regex.Replace(endeme, "(...)$", "-$1");
            if (divs >= 5 && len >= 18) endeme = Regex.Replace(endeme, "^(.............)", "$1-");

            return endeme.PadRight(width);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- CutChars -->
        /// <summary>
        ///      What is this supposed to do?
        /// </summary>
        /// <param name="endeme"></param>
        /// <param name="num"></param>
        /// <returns></returns>
        private string CutChars(string endeme, int num)
        {
            return endeme;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomLetter -->
        /// <summary>
        ///      Picks a letter, the first being th most likely
        /// </summary>
        /// <param name="geom"></param>
        /// <returns>change this to use the new approach</returns>
        public char RandomLetter(double geometric)
        {
            Random r = EndemeRandomSource.New().Random;
            char letter = ' ';
            for (int i = 0; i < _string.Count && letter == ' '; ++i)
                if (r.NextDouble() > geometric)
                    letter = _string[i];
            if (letter == ' ')
            {
                double num = Math.Abs(r.Next(6) + r.Next(6) + r.Next(6) + r.Next(6) - 8.5) - 0.5;
                if ((int)num < _string.Count)
                    letter = _string[(int)num];
            }
            return letter;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomLetter -->
        /// <summary>
        ///      Picks a letter, the first being th most likely
        /// </summary>
        /// <param name="level">ranges from 2:1 to 12:22</param>
        /// <returns>change this to use the new approach</returns>
        /// <remarks>
        ///       1  2  1/22 * (49/36)^10
        ///       2   
        ///       3  3  1/22 * (49/36)^9
        ///       4  4  1/22 * (49/36)^8
        ///       5   
        ///       6  5  1/22 * (49/36)^7
        ///       7   
        ///       8  6  1/22 * (49/36)^6
        ///       9   
        ///      10  7  1/22 * (49/36)^5
        ///      11   
        ///      12    
        ///      13  8  1/22 * (49/36)^4
        ///      14   
        ///      15  9  1/22 * (49/36)^3
        ///      16   
        ///      17   
        ///      18  10  1/22 * (49/36)^2
        ///      19    
        ///      20  11  1/22 * (49/36)
        ///      21    
        ///      22  12  1/22
        /// </remarks>
        public char RandomLetter(int end)
        {
            Random r = EndemeRandomSource.New().Random;
            char letter = ' ';

            double power = (22.0 - end) / 2.0;
            double factor = Math.Pow(49.0 / 36.0, power);
            double naive = factor / 22;

            double remaining = 1.0;
            double adjusted = naive;
            double percent = 0.0;


            int num = _string.Count;
            while (num >= _string.Count)
            {
                num = _string.Count;
                for (int position = 0; num >= _string.Count && position < 22; ++position)
                {
                    remaining = remaining - percent;
                    adjusted = 22 * naive / (22 - position);
                    double ran = r.Next(10000);
                    if (ran <= adjusted * 10000)
                        num = position;
                }
            }


            if ((int)num < _string.Count)
                letter = _string[(int)num];
            return letter;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StdDeviation -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="dictionary"></param>
        /// <param name="c"></param>
        /// <returns></returns>
        public static double StdDeviation(Dictionary<char, Endeme> dictionary, char c)
        {
            List<double> list = (from d in dictionary
                        select d.Value.Quant.Raw[c]).ToList();
            return StdDeviation(list);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StdDeviation -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="list"></param>
        /// <returns></returns>
        public static double StdDeviation(List<double> list)
        {
            double average = list.Average();
            double sumOfDifferencesSquared = list.Select(val => ((double)val - average) * ((double)val - average)).Sum();
            double sd = Math.Sqrt(sumOfDifferencesSquared / list.Count);
            return sd;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- ToList -->
        /// <summary>
        ///      Returns a copy of the list of characters
        /// </summary>
        /// <returns></returns>
        public List<char> ToList()
        {
            char[] cha = new char[_string.Count];
            this._string.CopyTo(cha);
            List<char> list = new List<char>(cha);
            return list;
        }


        // --------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string ToString(bool longFormat)
        {
            if (longFormat)
            {
                string setName;

                if (IsEmpty) return "empty";
                if (_set == null) setName = "[no set]"; else setName = _set.Name;
                if (string.IsNullOrEmpty(setName)) setName = "[unnamed set]";
                return (setName + ":" + ToString());
            }
            else
                return ToString();
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        ///      Returns a straight string or blank
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            if (IsEmpty || _string == null) return "";
            else return (new string(_string.ToArray()));
        }



        // ----------------------------------------------------------------------------------------
        /// <!-- Show() -->
        /// <summary>
        ///      Returns a string displaying the N most important characteristics etc
        /// </summary>
        /// <remarks>
        ///      Deprecated
        /// </remarks>
        /// <param name="num"></param>
        /// <param name="width"></param>
        /// <param name="wordLen"></param>
        /// <returns></returns>
        public string Show(int num, int width, string divider, int wordLen, string delim)
        {
            if (num > width) num = width;
            string str = "";
            string pattern = "";
            string replace = "";
            StringBuilder letters = new StringBuilder();


            if (IsEmpty || _string == null) str = "";
            else
            {
                str = new string(_string.ToArray());
                num = Math.Max(Math.Min(str.Length, num), 0);
                switch(width - num)
                {
                    case 0 :
                        letters = new StringBuilder(str);
                        break;
                    case 1 : case 2 :
                        if (num > 5)
                             letters.Append(Regex.Replace(str, "^(...)(..)", "$1-$2"));
                        else letters.Append(SetLength(num, str));
                        break;
                    case 3 :
                        switch(num)
                        {
                            case  0 : pattern = "^.*$"                                  ; replace = ""           ; break;
                            case  1 : pattern = "^(.).*$"                               ; replace = "$1"         ; break;
                            case  2 : pattern = "^(..).*$"                              ; replace = "$1"         ; break;
                            case  3 : pattern = "^(...).*$"                             ; replace = "$1"         ; break;
                            case  4 : pattern = "^(....).*$"                            ; replace = "$1"         ; break;
                            case  5 : pattern = "^(...)(..).*$"                         ; replace = "$1-$2"      ; break;
                            case  6 : pattern = "^(...)(...).*$"                        ; replace = "$1-$2"      ; break;
                            case  7 : pattern = "^(...)(....).*$"                       ; replace = "$1-$2"      ; break;
                            case  8 : pattern = "^(...)(.....).*$"                      ; replace = "$1-$2"      ; break;
                            case  9 : pattern = "^(...)(...).*(...)$"                   ; replace = "$1-$2--$3"  ; break;
                            case 10 : pattern = "^(...)(....).*(...)$"                  ; replace = "$1-$2--$3"  ; break;
                            case 11 : pattern = "^(...)(.....).*(...)$"                 ; replace = "$1-$2--$3"  ; break;
                            case 12 : pattern = "^(...)(....)(..).*(...)$"              ; replace = "$1-$2-$3-$4"; break;
                            case 13 : pattern = "^(...)(....)(...).*(...)$"             ; replace = "$1-$2-$3-$4"; break;
                            case 14 : pattern = "^(...)(....)(....).*(...)$"            ; replace = "$1-$2-$3-$4"; break;
                            case 15 : pattern = "^(...)(....)(....).*(....)$"           ; replace = "$1-$2-$3-$4"; break;
                            case 16 : pattern = "^(...)(....)(....).*(.....)$"          ; replace = "$1-$2-$3-$4"; break;
                            case 17 : pattern = "^(...)(....)(....).*(......)$"         ; replace = "$1-$2-$3-$4"; break;
                            case 18 : pattern = "^(...)(....)(....).*(.......)$"        ; replace = "$1-$2-$3-$4"; break;
                            case 19 : pattern = "^(...)(....)(....).*(........)$"       ; replace = "$1-$2-$3-$4"; break;
                            case 20 : pattern = "^(...)(....)(......).*(.......)$"      ; replace = "$1-$2-$3-$4"; break;
                            case 21 : pattern = "^(...)(....)(.......).*(.......)$"     ; replace = "$1-$2-$3-$4"; break;
                            case 22 : pattern = "^(...)(....)(........).*(.......)$"    ; replace = "$1-$2-$3-$4"; break;
                            default : pattern = "^(...)(....)(.......*)(.......)$"      ; replace = "$1-$2-$3-$4"; break;
                        }
                        letters.Append(Regex.Replace(str, pattern, replace));
                        break;
                  //  case 4 :
                  //      switch(num)
                  //      {
                  //          case  0 : pattern = "^.*$"                                  ; replace = ""              ; break;
                  //          case  1 : pattern = "^(.).*$"                               ; replace = "$1"            ; break;
                  //          case  2 : pattern = "^(..).*$"                              ; replace = "$1"            ; break;
                  //          case  3 : pattern = "^(...).*$"                             ; replace = "$1"            ; break;
                  //          case  4 : pattern = "^(....).*$"                            ; replace = "$1"            ; break;
                  //          case  5 : pattern = "^(...)(..).*$"                         ; replace = "$1-$2"         ; break;
                  //          case  6 : pattern = "^(...)(...).*$"                        ; replace = "$1-$2"         ; break;
                  //          case  7 : pattern = "^(...)(....).*$"                       ; replace = "$1-$2"         ; break;
                  //          case  8 : pattern = "^(...)(.....).*$"                      ; replace = "$1-$2"         ; break;
                  //          case  9 : pattern = "^(...)(...).*(...)$"                   ; replace = "$1-$2---$3"    ; break;
                  //          case 10 : pattern = "^(...)(....).*(...)$"                  ; replace = "$1-$2---$3"    ; break;
                  ///* TAG */ case 11 : pattern = "^(...)(.....).*(...)$"                 ; replace = "$1-$2---$3"    ; break;
                  //          case 12 : pattern = "^(...)(....)(..).*(...)$"              ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 13 : pattern = "^(...)(....)(...).*(...)$"             ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 14 : pattern = "^(...)(....)(....).*(...)$"            ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 15 : pattern = "^(...)(....)(....).*(....)$"           ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 16 : pattern = "^(...)(....)(....).*(.....)$"          ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 17 : pattern = "^(...)(....)(....).*(......)$"         ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 18 : pattern = "^(...)(....)(....).*(.......)$"        ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 19 : pattern = "^(...)(....)(....).*(........)$"       ; replace = "$1-$2-$3--$4"  ; break;
                  //          case 20 : pattern = "^(...)(....)(......).*(....)(...)$"    ; replace = "$1-$2-$3-$4-$5"; break;
                  //          case 21 : pattern = "^(...)(....)(.......).*(....)(...)$"   ; replace = "$1-$2-$3-$4-$5"; break;
                  //          case 22 : pattern = "^(...)(....)(........).*(....)(...)$"  ; replace = "$1-$2-$3-$4-$5"; break;
                  //          default : pattern = "^(...)(....)(.......*)(....)(...)$"    ; replace = "$1-$2-$3-$4-$5"; break;
                  //      }
                  //      letters.Append(Regex.Replace(str, pattern, replace));
                  //      break;
                    case 5 : case 6 : case 7 : case 8 : case 9 : case 10 :
                        switch(num)
                        {
                            case  0 : pattern = "^.*$"                                  ; replace = ""                 ; break;
                            case  1 : pattern = "^(.).*$"                               ; replace = "$1"               ; break;
                            case  2 : pattern = "^(..).*$"                              ; replace = "$1"               ; break;
                            case  3 : pattern = "^(...).*$"                             ; replace = "$1"               ; break;
                            case  4 : pattern = "^(....).*$"                            ; replace = "$1"               ; break;
                            case  5 : pattern = "^(...)(..).*$"                         ; replace = "$1-$2"            ; break;
                            case  6 : pattern = "^(...)(...).*$"                        ; replace = "$1-$2"            ; break;
                            case  7 : pattern = "^(...)(....).*$"                       ; replace = "$1-$2"            ; break;
                            case  8 : pattern = "^(...)(.....).*$"                      ; replace = "$1-$2"            ; break;
                            case  9 : pattern = "^(...)(...).*(...)$"                   ; replace = "$1-$2----$3"      ; break;
                            case 10 : pattern = "^(...)(....).*(...)$"                  ; replace = "$1-$2----$3"      ; break;
                            case 11 : pattern = "^(...)(.....).*(...)$"                 ; replace = "$1-$2----$3"      ; break;
                            case 12 : pattern = "^(...)(....)(..).*(...)$"              ; replace = "$1-$2-$3---$4"    ; break;
                            case 13 : pattern = "^(...)(....)(...).*(...)$"             ; replace = "$1-$2-$3---$4"    ; break;
                            case 14 : pattern = "^(...)(....)(....).*(...)$"            ; replace = "$1-$2-$3---$4"    ; break;
                            case 15 : pattern = "^(...)(....)(....).*(....)$"           ; replace = "$1-$2-$3---$4"    ; break;
                            case 16 : pattern = "^(...)(....)(....).*(..)(...)$"        ; replace = "$1-$2-$3--$4-$5"  ; break;
                            case 17 : pattern = "^(...)(....)(....).*(...)(...)$"       ; replace = "$1-$2-$3--$4-$5"  ; break;
                            case 18 : pattern = "^(...)(....)(....).*(....)(...)$"      ; replace = "$1-$2-$3--$4-$5"  ; break;
                            case 19 : pattern = "^(...)(....)(....).*(.....)(...)$"     ; replace = "$1-$2-$3--$4-$5"  ; break;
                            case 20 : pattern = "^(...)(....)(....)(..).*(....)(...)$"  ; replace = "$1-$2-$3-$4-$5-$6"; break;
                            case 21 : pattern = "^(...)(....)(....)(...).*(....)(...)$" ; replace = "$1-$2-$3-$4-$5-$6"; break;
                            case 22 : pattern = "^(...)(....)(....)(....).*(....)(...)$"; replace = "$1-$2-$3-$4-$5-$6"; break;
                            default : pattern = "^(...)(....)(....)(...*)(....)(...)$"  ; replace = "$1-$2-$3-$4-$5-$6"; break;
                        }
                        letters.Append(Regex.Replace(str, pattern, replace));
                        break;
                    case 4 :
                    default : // TAG
                        switch(num)
                        {
                            case  0 : pattern = "^.*$"                                  ; replace = ""              ; break;
                            case  1 : pattern = "^(.).*$"                               ; replace = "$1"            ; break;
                            case  2 : pattern = "^(..).*$"                              ; replace = "$1"            ; break;
                            case  3 : pattern = "^(...).*$"                             ; replace = "$1"            ; break;
                            case  4 : pattern = "^(....).*$"                            ; replace = "$1"            ; break;
                            case  5 : pattern = "^(...)(..).*$"                         ; replace = "$1-$2"         ; break;
                            case  6 : pattern = "^(...)(...).*$"                        ; replace = "$1-$2"         ; break;
                            case  7 : pattern = "^(...)(....).*$"                       ; replace = "$1-$2"         ; break;
                            case  8 : pattern = "^(...)(.....).*$"                      ; replace = "$1-$2"         ; break;
                            case  9 : pattern = "^(...)(...).*(...)$"                   ; replace = "$1-$2--$3"     ; break;
                            case 10 : pattern = "^(...)(....).*(...)$"                  ; replace = "$1-$2--$3"     ; break;
                  /* TAG */ case 11 : pattern = "^(...)(.....).*(...)$"                 ; replace = "$1-$2--$3"     ; break;
                            case 12 : pattern = "^(...)(....)(..).*(...)$"              ; replace = "$1-$2-$3--$4"  ; break;
                            case 13 : pattern = "^(...)(....)(...).*(...)$"             ; replace = "$1-$2-$3--$4"  ; break;
                            case 14 : pattern = "^(...)(....)(....).*(...)$"            ; replace = "$1-$2-$3--$4"  ; break;
                            case 15 : pattern = "^(...)(....)(....).*(....)$"           ; replace = "$1-$2-$3--$4"  ; break;
                            case 16 : pattern = "^(...)(....)(....).*(.....)$"          ; replace = "$1-$2-$3--$4"  ; break;
                            case 17 : pattern = "^(...)(....)(....).*(......)$"         ; replace = "$1-$2-$3--$4"  ; break;
                            case 18 : pattern = "^(...)(....)(....).*(.......)$"        ; replace = "$1-$2-$3--$4"  ; break;
                            case 19 : pattern = "^(...)(....)(....).*(........)$"       ; replace = "$1-$2-$3--$4"  ; break;
                            case 20 : pattern = "^(...)(....)(......).*(....)(...)$"    ; replace = "$1-$2-$3-$4-$5"; break;
                            case 21 : pattern = "^(...)(....)(.......).*(....)(...)$"   ; replace = "$1-$2-$3-$4-$5"; break;
                            case 22 : pattern = "^(...)(....)(........).*(....)(...)$"  ; replace = "$1-$2-$3-$4-$5"; break;
                            default : pattern = "^(...)(....)(.......*)(....)(...)$"    ; replace = "$1-$2-$3-$4-$5"; break;
                        }
                        letters.Append(Regex.Replace(str, pattern, replace));
                        break;
                }
            }


            // -------------------------------------------------------------------------
            //  Add some words if there is room
            // -------------------------------------------------------------------------
            if (_set != null && width > letters.Length + 5)
            {
                string d = divider;
                for (int i = 0; i < _string.Count && i < 7; ++i)
                {
                    letters.Append(d + SetLength(wordLen, _set.Characteristic[_string[i]].Label));
                    d = delim;
                }
                d = divider;
                for (int i = Math.Max(0, _string.Count-3); i < _string.Count; ++i)
                {
                    char ch = _string[i];
                    string addme = _set.Characteristic[ch].Label;
                    addme = Regex.Replace(addme, "^.* vs ", "-");
                    letters.Append(d + SetLength(wordLen, addme));
                    d = delim;
                }
                d = divider;
                for (int i = 7; i < _string.Count-3; ++i)
                {
                    letters.Append(d + SetLength(wordLen, _set.Characteristic[_string[i]].Label));
                    d = delim;
                }
            }
            return SetLength(width, letters.ToString());
        }

    }
}


