﻿// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;

namespace InformationLib.Endemes
{
	// THIS NAMESPACE IS A PRIMITIVE!  use only System.* references

	// --------------------------------------------------------------------------------------------
	/// <!-- EndemeAbstractComponent -->
	/// <summary>
	///      The EndemeAbstractComponent class is the component in the component pattern,
	///      the leaf is called EndemeItem, the composite is called EndemeNode
	/// </summary>
	/// <remarks>
	///      I have decided to use a modified form of the composite pattern:
	///      Here is the entire structure:
	/// 
	///                                        +-----------------------------+
	///                                        v                             |
	///                                   EndemeNode (composite)             |   (endeme info item)
	///                                        |                             |
	///                          int/char ---> |                             |
	///                                       /|\                            |
	///                               EndemeNodeConnector                    |
	///                                        |                             |
	///                    EndemeSet --------> |                             |
	///                     |  ^  ^            |                             |
	///                     |  |  |           /|\                            |
	///                     |  |  +------ EndemeList                         |
	///                     |  |               |                             |
	///                     |  |      label -> |                             |
	///                     |  |      idx ---> |   EndemeAbstractComponent   |
	///                     |  |      Guid --> |      / (component)  \       |
	///                     |  |              /|\    /                \      |
	///                     |  |    (leaf) EndemeItem - - - - - - - - EndemeNode (composite)
	///                     |  |               |                            
	///                     |  |               +---------+-------+------+------+
	///                     |  |               |         |       |      |      |
	///                     |  +------------ Endeme   Number   Text   Value   Guid (is this the same guid as above?)
	///                     |                  |                        |     
	///                     | (---- char ----) |                    (anything)
	///                    /|\                /|\                             
	///       EndemeCharacteristic ==== EndemeQuantification                  
	///                     |
	///         position -> |
	///                    /|\
	///                EndemeValue
	/// 
	/// 
	///      TODO:
	/// 
	///             knowledge database access
	///      [PART]  - direct                          
	///              - indirect
	///              - excel
	///      [PART]  - auto-construction
	///              - auto-populating
	///              - osmosis
	/// 
	///             paths
	///              - lower level endeme sets
	///              - higher level endeme sets
	///              - hierarchical paths
	///              - relational paths
	/// 
	///             level converters
	///              - EndemeSet to EndemeList
	///              - EndemeList to EndemeSet
	///              - valued characteristic to EndemeSet
	/// 
	///             fuzzy converters
	///              - EndemeNodeConnector to Fuzzylist
	///              - FuzzyList to EndemeNodeConnector
	///              - EndemeValue to FuzzyTerm
	///              - FuzzyTerm to EndemeValue
	///              - EndemeCharacteristic to FuzzyMeasure
	///              - FuzzyMeasure to EndemeCharacteristic
	///              - EndemeList to FuzzyList
	///              - EndemeSet to FuzzyList
	/// 
	///             non-info
	///              - RichSqlCommand factory generate stored procedure test strings
	///              - install old test code
	/// 
	/// </remarks>
	public abstract class EndemeAbstractComponent
	{

		// ----------------------------------------------------------------------------------------
		//  Members
		// ----------------------------------------------------------------------------------------
		public abstract Endeme Endeme { get; set; }		/// <summary>The label of the node or leaf</summary>
		public abstract string Label  { get; set; }		/// <summary>The Guid key of the node or leaf</summary>
		public abstract Guid   Key    { get; set; }
		public abstract object Value  { get; set; }
		public abstract double Number { get; set; }
		public abstract string Text   { get; set; }


		// ----------------------------------------------------------------------------------------
		/// <!-- Copy -->
		/// <summary>
		///      Returns a copy of the EndemeAbstractComponent
		/// </summary>
		/// <returns></returns>
		public abstract EndemeAbstractComponent Copy();


		// ----------------------------------------------------------------------------------------
		/// <!-- Count -->
		/// <summary>
		///      Returns the count of leaves
		/// </summary>
		/// <value></value>
		/// <returns></returns>
		/// <remarks></remarks>
		public abstract int Count { get; }


		// ----------------------------------------------------------------------------------------
		/// <!-- IsEmpty -->
		/// <summary>
		///      Determines whether the component is empty
		/// </summary>
		/// <value></value>
		/// <returns></returns>
		/// <remarks></remarks>
		public abstract bool IsEmpty { get; }

    }
}