// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;                 // for Dictionary, List
using System.Linq;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeList -->
    /// <summary>
    ///      The EndemeList class contains a Guid keyed, labeled, ordered, label indexed list of endemes
    /// </summary>
    /// <remarks>
    ///      An endeme list is managed a bit like a two-column special purpose database table with the label
    ///      put into the EndemeItem class
    /// 
    ///                                        +-----------------------------+
    ///                                        v                             |
    ///                                   EndemeNode (composite)             |   (endeme info item)
    ///                                        |                             |
    ///                          int/char ---> |                             |
    ///                                       /|\                            |
    ///                               EndemeNodeConnector                    |
    ///                                        |                             |
    ///                    EndemeSet --------> |                             |
    ///                     |  ^  ^            |                             |
    ///                     |  |  |           /|\                            |
    ///                     |  |  +------ EndemeList                         |
    ///                     |  |               |                             |
    ///                     |  |      label -> |                             |
    ///                     |  |      idx ---> |   EndemeAbstractComponent   |
    ///                     |  |      Guid --> |      / (component)  \       |
    ///                     |  |              /|\    /                \      |
    ///                     |  |    (leaf) EndemeItem - - - - - - - - EndemeNode (composite)
    ///                     |  |               |                            
    ///                     |  |               +---------+-------+------+------+
    ///                     |  |               |         |       |      |      |
    ///                     |  +------------ Endeme   Number   Text   Value   Guid (this is the same guid as above)
    ///                     |                  |                        |     
    ///                     | (---- char ----> |                    (anything)
    ///                    /|\                /|\                             
    ///       EndemeCharacteristic ==== EndemeQuantification                  
    ///                     |
    ///         position -> |
    ///                    /|\
    ///                EndemeValue
    /// 
    /// 
    /// 
    ///      Parallel Levels of Complexity Among Various Information Structures:
    /// 
    /// 
    ///      (EndemeNode)                  (?)
    ///         |                           |
    ///      (EndemeNodeConnector) = = = (FuzzyList)
    ///         |                           |
    ///      (EndemeList)                   |
    ///         |                           |
    ///         |                           |
    ///      EndemeSet, (Endeme)            |
    ///      (EndemeItem)                   |
    ///         |                        FuzzyMeasure, 
    ///         |                  = ? = (FuzzyItem)
    ///      EndemeCharacteristic           |
    ///         |                           |
    ///         |                        FuzzyTerm
    ///         |                  = ? =    
    ///      EndemeValue                    
    /// 
    /// </remarks>
    public class EndemeList
    {
        // ----------------------------------------------------------------------------------------
        //  Members
        // ----------------------------------------------------------------------------------------
        private List<EndemeAbstractComponent>               _list; // composite components?
        public  EndemeSet                      Set { get; set; }  // the endeme set for the endemes in the list
        private Dictionary<string, List<Guid>> _labelIndex; // index on the 'label' 'column'


        // ----------------------------------------------------------------------------------------
        //  Accessors
        // ----------------------------------------------------------------------------------------
        //public Endeme this [EndemeLeaf_old item] { get { return _endeme    [item.Key]; } set { _endeme[item.Key] = value; } } // I don't know if the set is a good idea
        public EndemeAbstractComponent this[Guid key] { get { return GetEndemeItem(key); } } /* set { _list[key] = value; } */
        public List<EndemeItem> this[string key] { get { List<EndemeItem> list = (from c in _list where c.Label == key select (EndemeItem)c).ToList(); return list; } }
        public EndemeAbstractComponent this[int idx] { get { return _list[idx]; } }
        public EndemeAbstractComponent this[Endeme tgt] { get { return GetEndemeItem(tgt); } }
        //List<EndemeAbstractComponent> list = FindClosest_sequentialSearch(1, tgt);
            //if (list.Count > 0) return list[0];
            //else return null;

        private EndemeItem GetEndemeItem(Endeme tgt)
        {
            List<EndemeItem> list = FindClosest_sequentialSearch(1, tgt);
            if (list.Count > 0) return list[0];
            else return null;
        }
        private EndemeAbstractComponent GetEndemeItem(Guid key)
        {
            List<EndemeAbstractComponent> list = (from c in _list
                        where c.Key.ToString() == key.ToString() // ?
                        select c).ToList();
            if (list.Count > 0) return list[0];
            else return null;
        }


        // ----------------------------------------------------------------------------------------
        //  Constructors
        // ----------------------------------------------------------------------------------------
        private EndemeList(                               ) { Init(22      , null   ); }
        public  EndemeList(              EndemeSet listSet) { Init(22      , listSet); }
        public  EndemeList(int capacity, EndemeSet listSet) { Init(capacity, listSet); }

         
        // ----------------------------------------------------------------------------------------
        //  Short methods
        // ----------------------------------------------------------------------------------------
        public bool       ContainsKey(Guid   key  ) { return (this.EndemeItem0(key) != null); }
        public bool       ContainsKey(int    idx  ) { return (0 <= idx && idx < _list.Count); }
        public bool       ContainsKey(string lbl  ) { return (this[lbl].Count > 0);           }
        public EndemeAbstractComponent EndemeItem0(char c) { return EndemeItem0("" + c); }
        public EndemeAbstractComponent EndemeItem0(string label) { List<EndemeAbstractComponent> list = (from c in _list where c.Label == label select c).ToList(); if (list.Count > 0) return list[0]; else return null; }
        public EndemeAbstractComponent EndemeItem0(Guid key) { List<EndemeAbstractComponent> list = (from c in _list where c.Key.ToString() == key.ToString() select c).ToList(); if (list.Count > 0) return list[0]; else return null; }


        // ----------------------------------------------------------------------------------------
        //  Short properties
        // ----------------------------------------------------------------------------------------
        public int          Count   { get { return _list.Count;                                                          } }
        public List<Endeme> Endemes { get { List<Endeme> list = (from c in _list select c.Endeme).ToList(); return list; } }
        public List<Guid>   Keys    { get { List<Guid>   list = (from c in _list select c.Key   ).ToList(); return list; } }
        public List<string> Labels  { get { List<string> list = (from c in _list select c.Label ).ToList(); return list; } }
        public List<double> Numbers { get { List<double> list = (from c in _list select c.Number).ToList(); return list; } }
        public List<object> Objects { get { List<object> list = (from c in _list select c.Value ).ToList(); return list; } }
        public List<string> Strings { get { List<string> list = (from c in _list select c.Text  ).ToList(); return list; } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Add -->
        /// <summary>
        ///      Add labeled endeme to the endeme list
        /// </summary>
        /// <param name="key">enter an empty to make it random</param>
        /// <param name="label"></param>
        /// <param name="endeme">endeme to add</param>
        /// <param name="value">value to store</param>
        public EndemeItem Add(Guid key, string label, Endeme endeme, object value)
        {
            // --------------------------------------------------------------------------
            //  Preliminaries
            // --------------------------------------------------------------------------
            if (Set == null)
                throw new Exception("please set endeme set before adding to list");


            // --------------------------------------------------------------------------
            //  Add key and endeme to the various data structures
            // --------------------------------------------------------------------------
            if (key == null || key == Guid.Empty)
                key = Guid.NewGuid();
            EndemeItem item = new EndemeItem(endeme);
            item.Value = value;


            item.Key = key;
            item.Label = label;
            _list.Add(item);


            // --------------------------------------------------------------------------
            //  Add endeme to the indexes
            // --------------------------------------------------------------------------
            if (!_labelIndex.ContainsKey(item.Label))
                _labelIndex[item.Label] = new List<Guid>(1);
            _labelIndex[item.Label].Add(key);


            return item;
        }
        public EndemeAbstractComponent Add(Endeme endeme) { return Add(Guid.NewGuid(), "", endeme, null); }
        public EndemeAbstractComponent Add(Endeme endeme, object value) { return Add(Guid.NewGuid(), "", endeme, value); }
        public EndemeAbstractComponent Add(Guid key, Endeme endeme) { return Add(key, "", endeme, null); }
        public EndemeAbstractComponent Add(Guid key, string label, Endeme endeme) { return Add(key, label, endeme, null); }


        // ----------------------------------------------------------------------------------------
        /// <!-- Average -->
        /// <summary>
        ///      Returns an endeme
        /// </summary>
        /// <returns></returns>
        public Endeme Average()
        {
            Endeme average = new Endeme(Set);
            if (_list == null || _list.Count == 0) return average;
            foreach (char c in Set.Letters)
                average.Quant.Raw[c] = _list.Average(item => item.Endeme.Quant.Raw[c]);
            average.CookEndeme();
            return average;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Copy -->
        /// <summary>
        ///      Returns a copy of the endeme list with a reference to the Set
        /// </summary>
        /// <returns></returns>
        public EndemeList Copy()
        {
            EndemeList list = new EndemeList();
            list.Set = this.Set;  // maybe this should be a Copy() also
            for (int i = 0; i < _list.Count; ++i)
            {
                list._list.Add((EndemeItem)_list[i].Copy());
            }
            foreach (string key in _labelIndex.Keys)
            {
                List<Guid> g = _labelIndex[key];
                List<Guid> guids = new List<Guid>(g.Count);
                for (int i = 0; i < g.Count; ++i)
                    guids.Add(g[i]);
                list._labelIndex.Add(key, guids);
            }
            return list;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- EvinceItem -->
        /// <summary>
        ///      Returns the endeme specified by the endeme string, creating it if needed
        /// </summary>
        /// <param name="list"></param>
        /// <param name="endeme"></param>
        /// <returns></returns>
        public EndemeAbstractComponent EvinceItem(string endeme)
        {
            EndemeAbstractComponent item = ExactMatch(endeme);
            if (item == null)
                item = Add(Guid.NewGuid(), new Endeme(Set, endeme));
            return item;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ExactMatch -->
        /// <summary>
        ///      Returns the endeme with the exact profile input
        /// </summary>
        /// <param name="endeme"></param>
        /// <returns></returns>
        public EndemeAbstractComponent ExactMatch(string endeme)
        {
            EndemeAbstractComponent found = null;
            foreach (EndemeAbstractComponent item in _list)
                if (endeme == item.Endeme.ToString())
                    found = item;
            return found;
        }
        public EndemeAbstractComponent ExactMatch(Endeme tgt) { return ExactMatch(tgt.ToString()); }


        // ----------------------------------------------------------------------------------------
        /// <!-- FindClosest -->
        /// <summary>
        ///      At some point I may build an index insde C#
        /// </summary>
        /// <param name="tgtEndeme"></param>
        /// <returns></returns>
        public List<EndemeItem> FindClosest(int n, Endeme tgtEndeme)
        {
            List<EndemeItem> closest = null;
            //if (_index != null && _isIndexed)
            //     closest = _index.FindClosest(n, tgtEndeme);
            //else
            closest = this.FindClosest_sequentialSearch(n, tgtEndeme);
            return closest;
        }
        public EndemeAbstractComponent FindClosest(Endeme tgtEndeme)
        {
            EndemeAbstractComponent closest = null;
            //if (_index != null && _isIndexed)
            //     closest = _index.FindClosest(n, tgtEndeme);
            //else
            closest = this.FindClosest_sequentialSearch(1, tgtEndeme)[0];
            return closest;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- FindClosest_sequentialSearch -->
        /// <summary>
        ///      Matches every item in the list to find the closest item to the input endeme
        /// </summary>
        /// <param name="list"></param>
        /// <param name="tgtEndeme"></param>
        /// <returns></returns>
        internal List<EndemeItem> FindClosest_sequentialSearch(int n, Endeme tgtEndeme)
        {
            // --------------------------------------------------------------------------
            //  Calculate the match with every item in the list
            // --------------------------------------------------------------------------
            List<EndemeItem> all = new List<EndemeItem>(Count);
            foreach (EndemeItem listItem in _list)
            {
                int match = this.Set.RawMatch(tgtEndeme, listItem.Endeme);
                EndemeItem item = new EndemeItem(listItem);
                item.Match = match;
                all.Add(item);
            }


            // --------------------------------------------------------------------------
            //  Return the top n
            // --------------------------------------------------------------------------
            List<EndemeItem> output = (from c in all
                                       orderby c.Match descending
                                       select c
                                      ).ToList();

            if (n >= all.Count) return output;

            int num = output.Count - n;
            output.RemoveRange(n, num);
            return output;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Init -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="capacity"></param>
        /// <param name="set"></param>
        public void Init(int capacity, EndemeSet set)
        {
            _list = new List<EndemeAbstractComponent>(capacity);
            Set   = set;
            _labelIndex = new Dictionary<string,List<Guid>>(capacity);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Normalize -->
        /// <summary>
        ///      Normalizes using multiplying
        /// </summary>
        /// <param name="avgTgt">average to normalize to</param>
        /// <param name="stdTgt">target standard deviation</param>
        public void Normalize(EndemeBalancer normalizer)
        {
            char[] letters = Set.Letters;


            // --------------------------------------------------------------------------
            //  Normalize the raw data
            //   (go through each endeme item and change the raw data based on the normalizers) (test me)
            // --------------------------------------------------------------------------
            foreach (EndemeAbstractComponent item in _list)
            {
                Endeme endeme = item.Endeme;
                normalizer.Normalize(endeme);
                //endeme.Normalize(normalizer);

                //item.Endeme.CookRawData(avgTgt, stdTgt, avg, std);
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomFill -->
        /// <summary>
        ///      Fills the list with random endemes then creates an index for it
        /// </summary>
        /// <param name="count"></param>
        public void RandomFill(int count)
        {
            Random r = EndemeRandomSource.New().Random;
            for (int n = 0; n < count; ++n)
                Add(Guid.Empty, n.ToString(), Set.RandomEndeme(r));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Remove -->
        /// <summary>
        ///      Removes an item from the list and all of its indexes
        /// </summary>
        /// <param name="key"></param>
        public void Remove(Guid key)
        {
            EndemeAbstractComponent idx = this[key];
            _list.Remove(idx);
            _labelIndex[idx.Label].Remove(key);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ResolveIndex -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public int ResolveIndex(Guid key)
        {
            for (int i = 0; i < _list.Count; ++i )
                if (key == _list[i].Key)
                    return i;
            return -1;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- StdDeviation -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public Endeme StdDeviation()
        {
            Endeme std = new Endeme(Set);
            foreach (char c in Set.Letters)
            {
                List<double> list = (from d in _list
                        select d.Endeme.Quant.Raw[c]).ToList();
                std.Quant.Raw[c] = StdDeviation(list);
            }
            std.CookEndeme();
            return std;
        }
        public static double StdDeviation(List<double> list)
        {
            if (list == null || list.Count == 0) return 0.0;
            double average = list.Average();
            double sumOfDifferencesSquared = list.Select(val => ((double)val - average) * ((double)val - average)).Sum();
            double sd = Math.Sqrt(sumOfDifferencesSquared / list.Count);
            return sd;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "" + Count + " " + Set.Name + "s";
        }
    }
}
