// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Linq;
// (C) Jon Grover, 2012

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
	// --------------------------------------------------------------------------------------------
	/// <!-- EndemeNode -->
	/// <summary>
	///      The EndemeNode class contains a collection by
	///      1) int/char,
	///      2) endeme/context, and
	///      3) endeme set name
	///      and is used for storing soft coded variables, members, and fields,
	///      It can also be accessed using an EndemePath
	/// </summary>
	/// <remarks>
	///      I have decided to use a modified form of the composite pattern:
	/// 
	///                                        +-----------------------------+
	///                                        v                             |
	///                                   EndemeNode (composite)             |   (endeme info item)
	///                                        |                             |
	///                          int/char ---> |                             |
	///                                       /|\                            |
	///                               EndemeNodeConnector                    |
	///                                        |                             |
	///                    EndemeSet --------> |                             |
	///                     |  ^  ^            |                             |
	///                     |  |  |           /|\                            |
	///                     |  |  +------ EndemeList                         |
	///                     |  |               |                             |
	///                     |  |      label -> |                             |
	///                     |  |      idx ---> |   EndemeAbstractComponent   |
	///                     |  |      Guid --> |      / (component)  \       |
	///                     |  |              /|\    /                \      |
	///                     |  |    (leaf) EndemeItem - - - - - - - - EndemeNode (composite)
	///                     |  |               |                            
	///                     |  |               +---------+-------+------+------+
	///                     |  |               |         |       |      |      |
	///                     |  +------------ Endeme   Number   Text   Value   Guid (this is the same guid as above)
	///                     |                  |                        |     
	///                     | (---- char ----> |                    (anything)
	///                    /|\                /|\                             
	///       EndemeCharacteristic ==== EndemeQuantification                  
	///                     |
	///         position -> |
	///                    /|\
	///                EndemeValue
	/// 
	/// 
	///             node           leaf                        node        leaf
	///       i=2 set=[person] label=[physician]  ====>  j=2 set=[set] label=[endeme]  ====>  number=3
	/// </remarks>
	public class EndemeNode : EndemeAbstractComponent
	{

		// ----------------------------------------------------------------------------------------
		//  Members
		// ----------------------------------------------------------------------------------------
		public Dictionary<int, EndemeNodeConnector> _valueGroup {
			get { return m__valueGroup; }
			set { m__valueGroup = value; }
		}

		private Dictionary<int, EndemeNodeConnector> m__valueGroup;

		public override Endeme Endeme { get; set; }
		public override string Label { get; set; }
		/// <summary>The Guid key of the node or leaf</summary>
		public override Guid Key { get; set; }


		// ----------------------------------------------------------------------------------------
		//  Members for storing data as information
		// ----------------------------------------------------------------------------------------
		public override object Value  { get; set; }
		public override double Number { get; set; }
		public override string Text   { get; set; }


		// ----------------------------------------------------------------------------------------
		//  Accessors - custom indexers
		// ----------------------------------------------------------------------------------------
        public EndemeList this[char key, EndemeSet inputSet] { get { return _valueGroup[(int)key][inputSet.Name]; } }
        public EndemeList this[int  key, EndemeSet inputSet] { get { return _valueGroup[key][inputSet.Name]; } }


		// ----------------------------------------------------------------------------------------
		//  Constructor
		// ----------------------------------------------------------------------------------------
		// TODO: why is there an endeme set passed in here?
		public EndemeNode(int capacity, EndemeSet eSet)
		{
			_valueGroup = new Dictionary<int, EndemeNodeConnector>(capacity);
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- Average -->
		/// <summary>
		/// 
		/// </summary>
		/// <param name="endeme"></param>
		/// <param name="set"></param>
		/// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
		/// <returns></returns>
		public double Average(string endeme, EndemeSet inputSet, double closeness)
		{
			double total = 0.0;
            foreach (char key in _valueGroup.Keys)
            {
				Endeme e = From(key, endeme, inputSet, closeness).Endeme;
				total += From(key, endeme, inputSet, closeness).Number;
			}
			return (total / _valueGroup.Count);
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- Copy -->
		/// <summary>
		///      Returns a copy of the EndemeItem where the Value and the Set are references
		/// </summary>
		/// <returns></returns>
		public override EndemeAbstractComponent Copy()
		{
			EndemeNode node = new EndemeNode(_valueGroup.Count, null);
			node.Value = this.Value;
			node.Number = this.Number;
			node.Text = this.Text;

			foreach (int key in m__valueGroup.Keys) {
				_valueGroup.Add(key, m__valueGroup[key].Copy());
			}
			return node;
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- Count -->
		/// <summary>
		///      Returns the count of leaves
		/// </summary>
		/// <value></value>
		/// <returns></returns>
		/// <remarks></remarks>
		public override int Count {
			get {
				Int32 totalCount = 0;
				foreach (EndemeNodeConnector collection in m__valueGroup.Values) {
					totalCount = totalCount + collection.Count;
				}
				return totalCount;
			}
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- EduceConnector -->
		/// <summary>
		///      Returns the requested group of lists, creating it if it does not yet exist
		/// </summary>
		/// <param name="key"></param>
		/// <returns></returns>
		public EndemeNodeConnector EduceConnector(char key)
		{
			int i = (int)key;
			if (!_valueGroup.ContainsKey(i)) {
				_valueGroup.Add(i, new EndemeNodeConnector());
			}
			EndemeNodeConnector dictionary = _valueGroup[i];
			return dictionary;
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- Through -->
		/// <summary>
		///      Gets the exact data context, this may not be what you want if you DON'T want to create a new one
		/// </summary>
		/// <param name="key"></param>
		/// <param name="endeme"></param>
		/// <param name="set"></param>
		/// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
		/// <returns></returns>
		public EndemeItem Through(char key, string endeme, EndemeSet inputSet, double closeness)
		{
			EndemeNodeConnector @group = EduceConnector(key);
			EndemeList list = @group.EduceList(inputSet);
			EndemeItem item =(EndemeItem)list.EvinceItem(endeme);
			return item;
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- From -->
		/// <summary>
		///      Returns the best matching data context, this may not be what you want of you DO want to create a new one
		/// </summary>
		/// <remarks>
		///      To add - rarely return null, try a best context approximation, include maybe a match threshold
		/// </remarks>
		/// <param name="key"></param>
		/// <param name="endeme"></param>
		/// <param name="inputSet"></param>
		/// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
		/// <returns></returns>
		public EndemeItem From(char key, string endeme, EndemeSet inputSet, double closeness)
		{
			EndemeNodeConnector @group = EduceConnector(key);
			EndemeList list = @group.EduceList(inputSet);
			EndemeItem item = (EndemeItem)list[new Endeme(inputSet, endeme)];
			if (item != null) {
				return item;
			} else {
				return null;
			}
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- IsEmpty -->
		/// <summary>
		///      Determines whether the node is empty
		/// </summary>
		/// <value></value>
		/// <returns></returns>
		/// <remarks></remarks>
		public override bool IsEmpty {
			get {
				if (_valueGroup == null | _valueGroup.Count == 0) {
					return true;
				} else {
					return false;
				}
			}
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- StdDeviation -->
		/// <summary>
		/// 
		/// </summary>
		/// <param name="endeme"></param>
		/// <param name="inputSet"></param>
		/// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
		/// <returns></returns>
		public double StdDeviation(string endeme, EndemeSet inputSet, double closeness)
		{
			List<double> list = new List<double>();
			foreach (int i in _valueGroup.Keys) {
				char key = (char)i;
				list.Add(From(key, endeme, inputSet, closeness).Number);
			}
			return StdDeviation(list);
		}
		public static double StdDeviation(List<double> list)
		{
			if (list == null || list.Count == 0) {
				return 0.0;
			}
			double average = list.Average();
			double sumOfDifferencesSquared = list.Select(val => (Convert.ToDouble(val) - average) * (Convert.ToDouble(val) - average)).Sum();
			double sd = Math.Sqrt(sumOfDifferencesSquared / list.Count);
			return sd;
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- ToDictionary -->
		/// <summary>
		/// 
		/// </summary>
		/// <param name="endeme"></param>
		/// <param name="closeness">desired closeness, ranges from 0.0 to 1.0</param>
		/// <returns></returns>
		public Dictionary<char, int> ToDictionary(string endeme, EndemeSet inputSet, double closeness)
		{
			Dictionary<char, int> dictionary = new Dictionary<char, int>(_valueGroup.Count);
			foreach (int i in _valueGroup.Keys) {
				char key = (char)i;
				dictionary.Add(key, Convert.ToInt32(From(key, endeme, inputSet, closeness).Number));
			}
			return dictionary;
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- ToString -->
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			if ((string.IsNullOrEmpty(Label))) {
				return "" + _valueGroup.Count;
			} else {
				return Label + " node: " + _valueGroup.Count;
			}
		}

	}

}