// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Linq;

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
	// --------------------------------------------------------------------------------------------
	/// <!-- EndemeNode -->
	/// <summary>
	///      The EndemeNode class contains a number of endeme lists, one for each EndemeSet
	///      name, each list contains a loosely coupled group of fields/columns/members, this
	///      probably has some advantages over the tight coupling of fields in an object or table
	/// </summary>
	/// <remarks>
	///                                        +-----------------------------+
	///                                        v                             |
	///                                   EndemeNode (composite)             |   (endeme info item)
	///                                        |                             |
	///                          int/char ---> |                             |
	///                                       /|\                            |
	///                               EndemeNodeConnector                    |
	///                                        |                             |
	///                    EndemeSet --------> |                             |
	///                     |  ^  ^            |                             |
	///                     |  |  |           /|\                            |
	///                     |  |  +------ EndemeList                         |
	///                     |  |               |                             |
	///                     |  |      label -> |                             |
	///                     |  |      idx ---> |   EndemeAbstractComponent   |
	///                     |  |      Guid --> |      / (component)  \       |
	///                     |  |              /|\    /                \      |
	///                     |  |    (leaf) EndemeItem - - - - - - - - EndemeNode (composite)
	///                     |  |               |                            
	///                     |  |               +---------+-------+------+------+
	///                     |  |               |         |       |      |      |
	///                     |  +------------ Endeme   Number   Text   Value   Guid (is this the same guid as above?)
	///                     |                  |                        |     
	///                     | (---- char ----) |                    (anything)
	///                    /|\                /|\                             
	///       EndemeCharacteristic ==== EndemeQuantification                  
	///                     |
	///         position -> |
	///                    /|\
	///                EndemeValue
	/// 
	///       
	///                        node                                    node                    ? leaf
	///       i=2 set=[person] field=[physiciane]  ====>  j=2 set=[set] field=[endeme]  ====>  number=3
	/// 
	/// 
	/// 
	///      Parallel Levels of Complexity Among Various Information Structures:
	/// 
	/// 
	/// 
	///      (EndemeNode)                  (?)
	///         |                           |
	///      (EndemeNodeConnector) = = = (FuzzyList)
	///         |                           |
	///      (EndemeList)                   |
	///         |                           |
	///         |                           |
	///      EndemeSet, (Endeme)            |
	///      (EndemeItem)                   |
	///         |                        FuzzyMeasure, 
	///         |                  = ? = (FuzzyItem)
	///      EndemeCharacteristic           |
	///         |                           |
	///         |                        FuzzyTerm
	///         |                  = ? =    
	///      EndemeValue                    
	/// 
	/// </remarks>
	public class EndemeNodeConnector
	{
		// ----------------------------------------------------------------------------------------
		//  Members
		// ----------------------------------------------------------------------------------------

		private Dictionary<string, EndemeList> _dictionary;

		// ----------------------------------------------------------------------------------------
		//  Accessors
		// ----------------------------------------------------------------------------------------
		public EndemeList this[string inputSet] {
			get { return _dictionary[inputSet]; }
		}


		// ----------------------------------------------------------------------------------------
		//  Constructor
		// ----------------------------------------------------------------------------------------
		public EndemeNodeConnector()
		{
			_dictionary = new Dictionary<string, EndemeList>();
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- Copy -->
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		/// <remarks></remarks>
		public EndemeNodeConnector Copy()
		{
			EndemeNodeConnector output = new EndemeNodeConnector();
			foreach (string key in _dictionary.Keys) {
				output._dictionary.Add(key, _dictionary[key].Copy());
			}
			return output;
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- Count -->
		/// <summary>
		///      Returns the count of leaves
		/// </summary>
		/// <value></value>
		/// <returns></returns>
		/// <remarks></remarks>
		public int Count {
			get {
				Int32 totalCount = 0;
				foreach (EndemeList list in _dictionary.Values) {
					totalCount = totalCount + list.Count;
				}
				return totalCount;
			}
		}


		// ----------------------------------------------------------------------------------------
		/// <!-- EduceList -->
		/// <summary>
		///      Returns the requested list, creating it if it does not yet exist
		/// </summary>
		/// <param name="inputSet"></param>
		/// <returns></returns>
		public EndemeList EduceList(EndemeSet inputSet)
		{
			if (!_dictionary.ContainsKey(inputSet.Name)) {
				_dictionary.Add(inputSet.Name, new EndemeList(1, inputSet));
			}
			EndemeList list = _dictionary[inputSet.Name];
			return list;
		}

	}
}