// ------------------------------------------------------------------------------------------------
// This file is part of InformationLib.
//
// InformationLib is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// InformationLib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with InformationLib.  If not, see <http://www.gnu.org/licenses/>.
// ------------------------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;                                // for StringBuilder(2)
using System.Text.RegularExpressions;             // for Regex(1)

namespace InformationLib.Endemes // THIS NAMESPACE IS A PRIMITIVE!  use only System.* references
{
    // --------------------------------------------------------------------------------------------
    /// <!-- EndemeSet -->
    /// <summary>
    ///      The EndemeSet class is a combinable, orderable list of characteristics,
    ///      The endeme set itself has no order and is a 'type' for endemes,
    /// 
    ///      This namespace is a primitive so it should use nothing but System references
    /// </summary>
    public class EndemeSet
    {
        // ----------------------------------------------------------------------------------------
        //  Basics
        // ----------------------------------------------------------------------------------------
        public Dictionary<char, EndemeCharacteristic> Characteristic { get { return _char; } }
        private Dictionary<char, EndemeCharacteristic> _char;


        // ----------------------------------------------------------------------------------------
        //  For performance
        // ----------------------------------------------------------------------------------------
        private int _bestMatch;
        private int _worstMatch;


        // ----------------------------------------------------------------------------------------
        /// <!-- Cutoff, Name, Resource -->                                                                       /// <summary>The number of characteristics used for combinations</summary>
        public Guid   ID       { get { return _id;       } set { _id       = value; } } private Guid   _id;
        public string Name     { get { return _name;     } set { _name     = value; } } private string _name;     /// <summary>The resource of the endeme set - what the characetristics are balanced within</summary>
        /// <summary>The first endeme letter is position 0, cutoff is the number of letters to be looked at, so if cutoff is 0 that is assumed no cutoff, If cutoff is 1, then only look at the first letter: for (int letter = 0; letter < 1; ++letter) ... Endemes can have their own individual cutoffs, but generally in that case there is a specific cutoff characteristic</summary>
        public int    Cutoff   { get { return _cutoff;   } set { _cutoff   = value; } } private int    _cutoff;   /// <summary>The name of the endeme set, does it make sense to use this natural key?</summary>
        public string Resource { get { return _resource; } set { _resource = value; } } private string _resource;


        // ----------------------------------------------------------------------------------------
        //  Accessor
        // ----------------------------------------------------------------------------------------
        public EndemeCharacteristic this[char c] { get { if (_char.ContainsKey(c)) return _char[c]; return null; } }


        // ----------------------------------------------------------------------------------------
        //  Constructor
        // ----------------------------------------------------------------------------------------
        public EndemeSet()                                         { Init(Guid.Empty,     "",   "",       0);      }
        public EndemeSet(string name)                              { Init(Guid.NewGuid(), name, "",       0);      }
        public EndemeSet(Guid guid, string name)                   { Init(guid,           name, "",       0);      }
        public EndemeSet(string name, string resource)             { Init(Guid.NewGuid(), name, resource, 0);      }
        public EndemeSet(string name, string resource, int cutoff) { Init(Guid.NewGuid(), name, resource, cutoff); }


        // ----------------------------------------------------------------------------------------
        //  Short methods
        // ----------------------------------------------------------------------------------------
        public  static EndemeSet Empty   { get { return new EndemeSet();                                       } }        /// <summary>Returns an ordered list of the characteristic letters</summary>
        private List<char> LettersSorted { get { List<char> list = LettersUnsorted; list.Sort();  return list; } } /// <summary>Returns an unordered array of the characteristic letters</summary>
        public  char[]     Letters       { get { return LettersUnsorted.ToArray();                             } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Add -->
        /// <summary>
        ///      Adds a characteristic to the endeme set
        /// </summary>
        /// <param name="cha"></param>
        /// <param name="label"></param>
        /// <param name="descr"></param>
        public void Add(char cha, string label, string descr)
        {
            if (cha > 'Z')
                cha = (char)((int)cha - 32);  // convert to upper case
            _char.Add(cha, new EndemeCharacteristic(cha, label, descr));
        }
        public void Add(char cha, string label, string descr, string units, double value, char opposite)
        {
            if (cha > 'Z')
                cha = (char)((int)cha - 32);  // convert to upper case
            _char.Add(cha, new EndemeCharacteristic(cha, label, descr, units, value, opposite));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Characteristics -->
        /// <summary>
        /// 
        /// </summary>
        public List<EndemeCharacteristic> Characteristics { get
        {
            List<char> letters = LettersSorted;
            List<EndemeCharacteristic> list = new List<EndemeCharacteristic>();
            foreach (char c in letters)
                list.Add(_char[c]);
            return list;
        } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Count -->
        /// <summary>
        ///      count of characteristics
        /// </summary>
        public int Count
        {
            get { return _char.Count; }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Contains -->
        /// <summary>
        ///      Checks for characteristic
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public bool Contains(char key)
        {
            return _char.ContainsKey(key);
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Diff -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="eq"></param>
        /// <param name="e2"></param>
        /// <returns></returns>
        public int Diff(Endeme e1, Endeme e2)
        {
            int total = 0;
            foreach (char cha in _char.Keys)
                if (e1.Contains(cha) && e2.Contains(cha))
                    total += Math.Abs((e1.Index(cha)) - (e2.Index(cha)));
                else
                    total += 49;
            return total;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Init -->
        /// <summary>
        ///      Standard Initializer, covering all members
        /// </summary>
        /// <param name="name"></param>
        /// <param name="resource"></param>
        /// <param name="cutoff"></param>
        private void Init(Guid id, string name, string resource, int cutoff)
        {
            _id         = id;
            _name       = name;
            _cutoff     = cutoff;
            _resource   = resource;
            _char       = new Dictionary<char, EndemeCharacteristic>();
            _bestMatch  = 0;
            _worstMatch = 0;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Labels -->
        /// <summary>
        ///      Returns a common delimited string containing the labels of all the characteristics
        /// </summary>
        public string Labels()
        {
            string str = "";
            string delim = "";
            foreach (char key in _char.Keys)
                { str += delim + key + _char[key].Label;  delim = ", "; }
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Labels -->
        /// <summary>
        ///      Returns a common delimited string showing the labels of the first n characteristics
        /// </summary>
        /// <param name="endeme">string specifying a (partial) endeme</param>
        /// <returns></returns>
        public string Labels(string endeme)
        {
            string list = "";
            string delim = "";
            char[] arr = endeme.ToCharArray();
            foreach (char c in arr)
            {
                list += delim + Characteristic[c].Label;
                delim = ", ";
            }
            return list;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Letter -->
        /// <summary>
        ///      Returns the letter for the characteristic specified by the labelPattern
        /// </summary>
        /// <param name="labelPattern">the label or regex pattern for the label</param>
        /// <returns></returns>
        public char Letter(string labelPattern)
        {
            char got = ' ';
            foreach (char key in _char.Keys)
            {
                if (Regex.IsMatch(_char[key].Label, labelPattern))
                {
                    if (got == ' ')
                        got = key;
                    else
                        throw new ArgumentException("Pattern '" + labelPattern + "' matches multiple characteristics"
                            + " in endeme set " + _name);
                }
            }
            return got;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- LettersUnsorted -->
        /// <summary>
        ///      Returns an unordered list of the characteristic letters
        /// </summary>
        /// <returns></returns>
        private List<char> LettersUnsorted { get
        {
            List<char> list = new List<char>(_char.Count);
            foreach (char key in _char.Keys)
                list.Add(key);
            return list;
        } }


        // ----------------------------------------------------------------------------------------
        /// <!-- Match -->
        /// <summary>
        ///      Returns match strength with another Endeme
        /// </summary>
        /// <param name="e1"></param>
        /// <param name="e2"></param>
        /// <param name="positivity">generally a positive number, most often 14,
        ///      indicates how much the matching is weighted toward high characteristics rather than low,
        ///      a number = half of endeme set length weighs positive and negative equally</param>
        /// <returns></returns>
        public double Match(Endeme e1, Endeme e2)
        {
            if (e1 == null || e2 == null || e1.IsEmpty || e2.IsEmpty)
                return 0.0;
            if (_bestMatch == 0)
            {
                Endeme straight = this.Straight;
                _bestMatch  = RawMatch(straight, straight);
                _worstMatch = RawMatch(straight, straight.Reverse);
            }
            int actualMatch = RawMatch(e1, e2);
            double num = (double)(actualMatch - _worstMatch) / (double)(_bestMatch - _worstMatch);
            if (num < 0.0) throw new Exception("kaboom");
            return num;
        }
        /// <summary>If you don't have an endeme set you can use this one but it's generally not a good idea</summary>
        public static double Match(Endeme e1, Endeme e2, string alphabet)
        {
            if (e1 == null || e2 == null || e1.IsEmpty || e2.IsEmpty)
                return 0.0;
            Endeme straight = new Endeme(alphabet);
            int bestMatch   = RawMatch(straight, straight,         alphabet);
            int worstMatch  = RawMatch(straight, straight.Reverse, alphabet);
            int actualMatch = RawMatch(e1, e2, alphabet);
            double num = (double)(actualMatch - worstMatch) / (double)(bestMatch - worstMatch);
            if (num < 0.0) throw new Exception("kaboom");
            return num;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Match_simple -->
        /// <summary>
        ///      Returns a raw match
        /// </summary>
        /// <param name="endeme"></param>
        /// <param name="maxCharValue"></param>
        /// <returns></returns>
        public int RawMatch(Endeme e1, Endeme e2)
        {
            int total = 0;
            if (e1 != null && e2 != null)
            {
                foreach (char cha in _char.Keys)
                    if (e1.Contains(cha) && e2.Contains(cha))
                        total += (e1.Index(cha)) * (e2.Index(cha));
                    else
                        total += 49;
            }
            else
                Pause();
            return total;
        }
        public void Pause() { }
        public static int RawMatch(Endeme e1, Endeme e2, string alphabet)
        {
            int total = 0;
            char[] keys = alphabet.ToCharArray();
            foreach (char cha in keys)
                if (e1.Contains(cha) && e2.Contains(cha))
                    total += (e1.Index(cha)) * (e2.Index(cha));
                else
                    total += 49;
            return total;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomCharacteristic -->
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public char RandomCharacteristic()
        {
            Random r = EndemeRandomSource.New().Random;
            List<char> key = new List<char>(_char.Keys);
            return key[r.Next(key.Count)];
        }
        /// <summary>Returns a random characteristic label</summary>
        public string RandomLabel() { return _char[RandomCharacteristic()].Label; }


        // ----------------------------------------------------------------------------------------
        /// <!-- RandomEndeme -->
        /// <summary>
        ///      Returns a random endeme, part of this set
        /// </summary>
        /// <returns></returns>
        public Endeme RandomEndeme(Random r)
        {
            List<char> list = new List<char>(_char.Keys);
            int count = list.Count;
            StringBuilder endeme = new StringBuilder(count);
            for (int i = 0; i < count; ++i)
            {
                int n = r.Next(list.Count);
                endeme.Append(list[n]);
                list.RemoveAt(n);
            }
            return (new Endeme(this, endeme.ToString()));
        }
        public Endeme RandomEndeme() { return RandomEndeme(EndemeRandomSource.New().Random); }


        // ----------------------------------------------------------------------------------------
        /// <!-- Set -->
        /// <summary>
        ///      Adds a new characteristic or sets an existing one to the new value
        /// </summary>
        /// <param name="cha"></param>
        /// <param name="label"></param>
        /// <param name="descr"></param>
        public void Set(char cha, string label, string descr)
        {
            if (_char.ContainsKey(cha))
                _char[cha].Set(cha, label, descr);
            else
                _char.Add(cha, new EndemeCharacteristic(cha, label, descr));
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ShowDetails -->
        /// <summary>
        ///      Lists the details for the top num characteristics in the endeme
        /// </summary>
        /// <param name="num"></param>
        /// <param name="endeme"></param>
        public string ShowDetails(int num, Endeme endeme)
        {
            string str = "";
            string delim = "";
            for (int i = 0; i < num; ++i)
            {
                EndemeCharacteristic c = _char[endeme[i]];
                str += delim + c.Letter + ". " + c.Label.PadRight(22) + " " + c.Descr;
                delim = "\r\n";
            }
            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ShowCharacteristics -->
        /// <summary>
        /// 
        /// </summary>
        /// <param name="set"></param>
        /// <returns></returns>
        public string ShowCharacteristics()
        {
            List<EndemeCharacteristic> list = this.Characteristics;
            string delim = "";
            string str = "";
            int maxLen = 0;
            foreach (EndemeCharacteristic cha in list)
                maxLen = Math.Max(maxLen, cha.Label.Length);
            foreach (EndemeCharacteristic cha in list)
            {
                str += delim + cha.Letter + ". " + cha.Label.PadRight(maxLen) + " - " + cha.Descr;
                delim = "\r\n";
            }

            return str;
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- Straight -->
        /// <summary>
        ///      Returns an endeme using all of the endeme set characteristics in order
        /// </summary>
        public Endeme Straight
        {
            get
            {
                string str = "";
                foreach (char cha in _char.Keys)
                    str += cha;
                Endeme e = new Endeme(this, str);
                return e;
            }
        }


        // ----------------------------------------------------------------------------------------
        /// <!-- ToString -->
        /// <summary>
        ///      Returns the endeme set in 'standard' format
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            List<char> keys = LettersSorted;
            StringBuilder str = new StringBuilder(keys.Count * 10); // *Magic number* alert
            string delim = "";
            foreach (char cha in keys)
            {
                str.Append(delim + cha + ")" + _char[cha].Label);
                delim = ", ";
            }
            return str.ToString();
        }

    }
}
